from iqEditors.commons import images, getImage
    # на случай, если кому-то понадобятся картинки
from iqEditors.simpleEditor import iQEditor     # простой редактор
from iqEditors.noteEditor import iQNoteEditor   # редактор записок
from iqEditors.providers import HelpLayerProvider, HelpBookProvider # диалоги справок


